from __future__ import annotations

import uuid

from hashlib import md5
from typing import TYPE_CHECKING

from sqlalchemy import Boolean, Unicode, Uuid, desc, select
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import Mapped, Session, mapped_column, relationship

from ..db.base_class import reg
from .login_history import LoginHistory
from .user_role import UserRole


if TYPE_CHECKING:
    from .role import Role
    from .voucher import Voucher


def encrypt(val: str) -> str:
    return md5(val.encode("utf-8") + "Salt".encode("utf-8")).hexdigest()


@reg.mapped_as_dataclass(unsafe_hash=True)
class User:
    __tablename__ = "users"

    id: Mapped[uuid.UUID] = mapped_column(Uuid, primary_key=True, insert_default=uuid.uuid4)
    name: Mapped[str] = mapped_column("username", Unicode, unique=True, nullable=False)
    _password: Mapped[str] = mapped_column("password", Unicode, nullable=False)
    locked_out: Mapped[bool] = mapped_column("disabled", Boolean, nullable=False)

    roles: Mapped[list[Role]] = relationship(secondary=UserRole.__table__)  # type: ignore[attr-defined]
    login_history: Mapped[list[LoginHistory]] = relationship(order_by=desc(LoginHistory.date), back_populates="user")
    vouchers: Mapped[list[Voucher]] = relationship("Voucher", foreign_keys="Voucher.user_id", back_populates="user")

    @hybrid_property
    def password(self) -> str:
        return self._password

    @password.inplace.setter
    def _password_setter(self, value: str) -> None:
        self._password = encrypt(value)

    def __init__(self, name: str, password: str, locked_out: bool, id_: uuid.UUID | None = None) -> None:
        self.name = name
        self.password = password
        self.locked_out = locked_out
        if id_ is not None:
            self.id = id_

    @classmethod
    def auth(cls, name: str, password: str, db: Session) -> "User" | None:
        if password is None:
            return None
        user: User | None = db.execute(select(cls).where(cls.name.ilike(name))).scalars().one_or_none()
        if not user:
            return None
        if user.password != encrypt(password) or user.locked_out:
            return None
        else:
            return user
