import datetime

import brewman.schemas.batch as schemas

from fastapi import APIRouter, Depends
from sqlalchemy import select
from sqlalchemy.orm import contains_eager

from ..core.security import get_current_active_user as get_user
from ..db.session import SessionFuture
from ..models.batch import Batch
from ..models.product import Product
from ..models.stock_keeping_unit import StockKeepingUnit
from ..schemas.user import UserToken


router = APIRouter()


@router.get("", response_model=list[schemas.Batch])
def batch_term(
    q: str,
    d: str,
    current_user: UserToken = Depends(get_user),
) -> list[schemas.Batch]:
    date_ = datetime.datetime.strptime(d, "%d-%b-%Y")
    list_: list[schemas.Batch] = []
    with SessionFuture() as db:
        query = (
            select(Batch)
            .join(Batch.sku)
            .join(StockKeepingUnit.product)
            .where(Batch.quantity_remaining > 0, Batch.name <= date_)
            .options(contains_eager(Batch.sku).contains_eager(StockKeepingUnit.product))
        )
        if q is not None:
            for q_item in q.split():
                query = query.where(Product.name.ilike(f"%{q_item}%"))
        result = db.execute(query.order_by(Product.name).order_by(Batch.name)).scalars().all()

        for item in result:
            text = (
                f"{item.sku.product.name} ({item.sku.units}) {item.quantity_remaining:.2f}@"
                f"{item.rate:.2f} from {item.name.strftime('%d-%b-%Y')}"
            )
            list_.append(
                schemas.Batch(
                    id_=item.id,
                    name=text,
                    quantity_remaining=round(item.quantity_remaining, 2),
                    rate=round(item.rate, 2),
                    tax=round(item.tax, 5),
                    discount=round(item.discount, 5),
                    sku=schemas.ProductLink(
                        id_=item.sku.id,
                        name=f"{item.sku.product.name} ({item.sku.units})",
                    ),
                )
            )
        return list_
