from calendar import monthrange
from datetime import date, datetime
from decimal import Decimal

from fastapi import APIRouter, Body, HTTPException, Security, status
from sqlalchemy import or_, select
from sqlalchemy.orm import Session

from ..core.security import get_current_active_user as get_user
from ..core.session import get_first_day, get_last_day
from ..db.session import SessionFuture
from ..models.account import Account
from ..models.attendance import Attendance
from ..models.attendance_type import AttendanceType
from ..models.employee import Employee
from ..models.journal import Journal
from ..models.voucher import Voucher
from ..models.voucher_type import VoucherType
from ..schemas.message import Message
from ..schemas.user import UserToken


router = APIRouter()


@router.post("", response_model=Message)
def credit_salary(
    month: str = Body(..., embed=True),
    user: UserToken = Security(get_user, scopes=["attendance"]),
) -> Message:
    month_ = datetime.strptime(month, "%d-%b-%Y").date()
    start_date = get_first_day(month_)
    finish_date = get_last_day(month_)
    with SessionFuture() as db:
        voucher = Voucher(
            date_=finish_date,
            narration="Auto Generated Salary Entry",
            user_id=user.id_,
            voucher_type=VoucherType.JOURNAL,
            posted=True,
            poster_id=user.id_,
        )
        db.add(voucher)
        for item in salary_journals(start_date, finish_date, db):
            voucher.journals.append(item)
            db.add(item)
        db.commit()
        return Message(message="Salary Entry created")


def salary_journals(start_date: date, finish_date: date, db: Session) -> list[Journal]:
    days = monthrange(start_date.year, start_date.month)[1]
    amount = Decimal(0)
    journals: list[Journal] = []
    employees = (
        db.execute(
            select(Employee)
            .where(Employee.joining_date <= finish_date, or_(Employee.is_active, Employee.leaving_date >= start_date))
            .order_by(Employee.cost_centre_id, Employee.designation, Employee.name)
        )
        .scalars()
        .all()
    )

    for employee in employees:
        attendances = (
            db.execute(
                select(Attendance).where(
                    Attendance.employee_id == employee.id,
                    Attendance.date_ >= start_date,
                    Attendance.date_ <= finish_date,
                    Attendance.is_valid == True,  # noqa: E712
                )
            )
            .scalars()
            .all()
        )
        att: Decimal
        att = sum(AttendanceType.by_id(x.attendance_type).value for x in attendances)  # type: ignore[assignment]
        att = round(att * employee.salary / days)  # type: ignore[assignment]
        if att != 0:
            amount += att
            journals.append(
                Journal(
                    amount=att,
                    debit=-1,
                    account_id=employee.id,
                    cost_centre_id=employee.cost_centre_id,
                )
            )
    salary = db.execute(select(Account).where(Account.id == Account.salary_id())).scalar_one()
    if amount == 0:
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail="No salaries to credit",
        )
    journals.append(
        Journal(
            amount=amount,
            debit=1,
            account_id=salary.id,
            cost_centre_id=salary.cost_centre_id,
        )
    )
    return journals
