import uuid

from datetime import date, datetime

from pydantic import (
    BaseModel,
    ConfigDict,
    FieldSerializationInfo,
    field_serializer,
    field_validator,
)

from . import to_camel
from .account import AccountLink
from .rate_contract_item import RateContractItem
from .user_link import UserLink


class RateContractBlank(BaseModel):
    date_: date
    valid_from: date | None
    valid_till: date | None
    narration: str
    items: list[RateContractItem]
    model_config = ConfigDict(str_strip_whitespace=True, alias_generator=to_camel, populate_by_name=True)

    @field_validator("date_", mode="before")
    @classmethod
    def parse_date(cls, value: date | str) -> date:
        if isinstance(value, date):
            return value
        return datetime.strptime(value, "%d-%b-%Y").date()

    @field_serializer("date_")
    def serialize_date(self, value: date, info: FieldSerializationInfo) -> str:
        return value.strftime("%d-%b-%Y")

    @field_validator("valid_from", mode="before")
    @classmethod
    def parse_valid_from(cls, value: date | str | None) -> date | None:
        if value is None:
            return value
        if isinstance(value, date):
            return value
        return datetime.strptime(value, "%d-%b-%Y").date()

    @field_serializer("valid_from")
    def serialize_valid_from(self, value: date | None, info: FieldSerializationInfo) -> str | None:
        return None if value is None else value.strftime("%d-%b-%Y")

    @field_validator("valid_till", mode="before")
    @classmethod
    def parse_valid_till(cls, value: date | str | None) -> date | None:
        if value is None:
            return value
        if isinstance(value, date):
            return value
        return datetime.strptime(value, "%d-%b-%Y").date()

    @field_serializer("valid_till")
    def serialize_valid_till(self, value: date | None, info: FieldSerializationInfo) -> str | None:
        return None if value is None else value.strftime("%d-%b-%Y")


class RateContractIn(RateContractBlank):
    vendor: AccountLink
    model_config = ConfigDict(str_strip_whitespace=True, alias_generator=to_camel, populate_by_name=True)

    @field_validator("date_", mode="before")
    @classmethod
    def parse_date(cls, value: date | str) -> date:
        if isinstance(value, date):
            return value
        return datetime.strptime(value, "%d-%b-%Y").date()


class RateContract(RateContractIn):
    id_: uuid.UUID
    creation_date: datetime
    last_edit_date: datetime
    user: UserLink
    model_config = ConfigDict(str_strip_whitespace=True, alias_generator=to_camel, populate_by_name=True)

    @field_validator("creation_date", mode="before")
    @classmethod
    def parse_creation_date(cls, value: datetime | str) -> datetime | None:
        if isinstance(value, datetime):
            return value
        return datetime.strptime(value, "%d-%b-%Y %H:%M")

    @field_serializer("creation_date")
    def serialize_creation_date(self, value: datetime, info: FieldSerializationInfo) -> str:
        return value.strftime("%d-%b-%Y %H:%M")

    @field_validator("last_edit_date", mode="before")
    @classmethod
    def parse_last_edit_date(cls, value: datetime | str) -> datetime | None:
        if isinstance(value, datetime):
            return value
        return datetime.strptime(value, "%d-%b-%Y %H:%M")

    @field_serializer("last_edit_date")
    def serialize_last_edit_date(self, value: datetime, info: FieldSerializationInfo) -> str:
        return value.strftime("%d-%b-%Y %H:%M")
